#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################################
# Usage: Dr.Web CommuniGate Plugin [-h] [-v] [-d BASE_DIR] [-r RSPAMD_HTTP_SOCKET] [-u RSPAMD_UNIX_SOCKET] [--debug]
#
# optional arguments:
# -h, --help show this help message and exit
# -v, --version         Show plugin version and exit
# -d, --base_dir CommuniGate base directory (default: /var/CommuniGate)
# -r, --rspamd_http_socket RspamD http socket address
# -u, --rspamd_unix_socket RspamD unix socket path
# -t, --rspamd_request_timeout RspamD request timeout (default: 120 sec). Value must be set in seconds
# -p, --plugin_directory Plugin files directory (default: /var/CommuniGate/DrWebPlugin)
# --debug Enable debug mode
#######################################################################################


import argparse
import asyncio
import io
import json
import logging
import math
import os
import re
import sys
import threading
from abc import abstractmethod
from asyncio import BaseTransport, Future, StreamReader, StreamWriter, Task, WriteTransport, get_event_loop
from collections import deque
from contextlib import contextmanager
from copy import deepcopy
from logging.handlers import TimedRotatingFileHandler
from types import TracebackType
from typing import Any, AsyncIterable, Awaitable, BinaryIO, Callable, Collection, Deque, Dict, Generator, Iterable, \
    List, Mapping, Optional, Sequence, Tuple, Type, Union

__author__ = "Doctor Web"
__copyright__ = "© Doctor Web, 1992–2022"

__build = "2204261223"
__version__ = "11.1.4" + (f".{__build}" if __build.isdecimal() else "")


# utils
#######

_thread_local = threading.local()
setattr(_thread_local, "ctx", {})
if sys.version_info < (3, 7):
    _current_task = asyncio.Task.current_task
else:
    _current_task = asyncio.current_task


def get_current_task() -> Optional[Task]:
    try:
        return _current_task()
    except RuntimeError:
        return None


def _get_ctx() -> Dict[Any, Any]:
    current_task = get_current_task()
    task_ctx = _thread_local.ctx.get(current_task)
    if task_ctx is None:
        task_ctx = {}
        _thread_local.ctx[current_task] = task_ctx
    return task_ctx


def set_ctx(name: str, value: Any) -> None:
    ctx = _get_ctx()
    ctx.setdefault(name, value)


def get_current_ctx(name: str) -> Any:
    ctx = _get_ctx()
    return ctx.get(name)


def copy_ctx_to_task(task: Task) -> None:
    ctx = _get_ctx()
    _thread_local.ctx[task] = deepcopy(ctx)


def empty_ctx(task: Optional[Task] = None) -> None:
    if task is None:
        task = get_current_task()

    if task is None:
        return

    _thread_local.ctx.pop(task)


class Job:
    _logger = logging.getLogger("background")

    def __init__(
            self,
            coro_or_fut: Awaitable,
            callback: Optional[Callable[['Job', Any], Awaitable[None]]] = None,
            name: Optional[str] = None,
    ) -> None:
        self._main_task = get_event_loop().create_task(self._start(coro_or_fut))
        self._callback = callback
        self._name = name

    @property
    def name(self) -> Optional[str]:
        return self._name

    async def _start(self, coro: Awaitable) -> Any:
        if self._name:
            set_ctx("background_job", self._name)

        result: Optional[Any] = None
        try:
            result = await coro
        except asyncio.CancelledError:
            self._logger.debug("Background job was cancelled")
        except Exception:
            self._logger.exception("Background job:", exc_info=True)
        else:
            self._logger.debug(f"Background job result: {result}")
            return result
        finally:
            if self._callback:
                try:
                    await self._callback(self, result)
                except Exception:
                    self._logger.exception("Background job callback exception:", exc_info=True)
                else:
                    self._logger.debug("Background job callback finished")
            empty_ctx(self._main_task)

    def cancel(self) -> None:
        self._main_task.cancel()


# Drweb RSPAMD client
############

class RSPAMDConnectionError(Exception):
    pass


class RSPAMDResponseError(Exception):
    pass


class RSPAMDTimeoutError(Exception):
    pass


class CommunigateMessageError(Exception):
    pass


class Connection:
    def __init__(self, reader: StreamReader, writer: StreamWriter) -> None:
        self._reader = reader
        self._writer = writer

    async def close(self) -> None:
        self._writer.transport.close()

    async def write(self, data: bytes) -> None:
        self._writer.write(data)
        await self._writer.drain()

    def at_eof(self) -> bool:
        return self._reader.at_eof()

    async def read_chunk(self, size: Optional[int] = None) -> bytes:
        reader = self._reader
        if size:
            return await reader.read(size)

        lines: List[bytes] = []
        while True:
            line = await reader.readline()
            if reader.at_eof() or line == b"\r\n":
                break

            lines.append(line)
        return b"".join(lines)


class BaseConnector:
    _logger: logging.Logger

    def __init__(self, address: str) -> None:
        self._address = address
        self._connections: List[Connection] = []

    @property
    def address(self) -> str:
        return self._address

    def __init_subclass__(cls, logger: Optional[logging.Logger] = None) -> None:
        cls._logger = logging.root if logger is None else logger

    @abstractmethod
    async def _open_connection(self) -> Tuple[StreamReader, StreamWriter]:
        ...

    async def acquire(self) -> Connection:
        self._logger.debug(f"Connecting to {self._address}...")

        reader, writer = await self._open_connection()
        connection = Connection(reader, writer)
        self._connections.append(connection)
        return connection

    async def release(self, connection: Connection) -> None:
        self._connections.remove(connection)
        await connection.close()
        self._logger.debug("Closed connection")

    async def close(self) -> None:
        for connection in self._connections:
            await connection.close()
        self._connections = []


class TcpConnector(BaseConnector, logger=logging.getLogger("tcp_connector")):

    def __init__(self, host: str, port: Union[int, str]):
        super(TcpConnector, self).__init__(f"{host}:{port}")
        self._host = host
        self._port = int(port)

    async def _open_connection(self) -> Tuple[StreamReader, StreamWriter]:
        reader, writer = await asyncio.open_connection(self._host, self._port)
        sock_name = writer.transport.get_extra_info("socket").getsockname()
        self._logger.debug(f"Created connection {sock_name[0]}:{sock_name[1]}")
        return reader, writer


class UnixConnector(BaseConnector, logger=logging.getLogger("unix_connector")):
    async def _open_connection(self) -> Tuple[StreamReader, StreamWriter]:
        reader, writer = await asyncio.open_unix_connection(self._address)
        peer_name = writer.transport.get_extra_info("socket").getpeername()
        self._logger.debug(f"Created connection {peer_name}")
        return reader, writer


class ResponseError(Exception):
    pass


class HTTPRequest:
    _logger = logging.getLogger("client_session_request")

    def __init__(
            self,
            connection: Connection,
            method: str,
            url: str,
            headers: Iterable[Tuple[str, Optional[str]]],
            content_size: int,
            content_iterator: AsyncIterable,
    ) -> None:
        self._connection = connection
        self._method = method
        self._url = url
        self._headers = headers
        self._content_size = content_size
        self._content_iterator = content_iterator

    async def _send_request(self) -> None:
        connection = self._connection
        start_message = f"{self._method.upper()} {self._url} HTTP/1.1" \
            f"\r\nUser-Agent: Drweb ComminiGate Pro plugin" \
            f"\r\nContent-Length: {self._content_size}"

        headers = "\r\n".join(f"{name}:{val}" for name, val in (self._headers or ()) if val)
        if headers:
            start_message += "\r\n" + headers
        start_message += "\r\n\r\n"
        await connection.write(start_message.encode("utf8"))
        self._logger.debug(f"Request headers was sent: {start_message}")

        async for chunk in self._content_iterator:
            await connection.write(chunk)
        self._logger.debug("Request body was sent")

    async def _read_response(self) -> Tuple[bytes, Dict[str, bytes], bytes]:
        connection = self._connection
        headers_chunk = await connection.read_chunk()
        if not headers_chunk and connection.at_eof():
            raise ResponseError("Empty errors")

        headers_lines = headers_chunk.split(b"\r\n")
        status = headers_lines[0].split(b" ")[1]
        headers: Dict[str, bytes] = {}
        for line in headers_lines[1:]:
            if not line:
                continue
            header_name, *header_value = line.rstrip(b"\r\n").lstrip(b"\n").split(b":")
            headers[header_name.strip().decode("utf8")] = b":".join(header_value).strip()
        self._logger.debug(f"Response headers: {headers}")

        content = await connection.read_chunk(int(headers["Content-Length"]))
        self._logger.debug(f"Received response body: {content!r}")
        return status, headers, content

    async def __call__(self) -> Tuple[bytes, Dict[str, bytes], bytes]:
        self._logger.debug("Sending request...")
        await self._send_request()
        self._logger.debug("Waiting response...")
        return await self._read_response()


class ClientSession:
    def __init__(self, connector: BaseConnector, timeout: Optional[float] = None) -> None:
        self._connector = connector
        self._timeout = timeout

    def close(self) -> Awaitable[None]:
        return self._connector.close()

    async def request(
            self,
            method: str,
            url: str,
            headers: Iterable[Tuple[str, Optional[str]]],
            content_size: int,
            content_iterator: AsyncIterable,
            timeout: Optional[float] = None,
    ) -> Tuple[bytes, Dict[str, bytes], bytes]:
        timeout = timeout if timeout is not None else self._timeout
        connection = await self._connector.acquire()
        try:
            request = HTTPRequest(connection, method, url, headers, content_size, content_iterator)
            request_task = get_event_loop().create_task(request())
            copy_ctx_to_task(request_task)
            try:
                return await asyncio.wait_for(request_task, timeout=timeout)
            finally:
                empty_ctx(request_task)
        finally:
            await self._connector.release(connection)

    async def __aenter__(self) -> 'ClientSession':
        return self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType],
    ) -> None:
        await self.close()


class RSPAMDService:
    _logger = logging.getLogger("rspamd")

    def __init__(self, client_session: ClientSession) -> None:
        self._client_session = client_session

    @staticmethod
    def _prepare_check_file_result(content: Mapping[str, Any]) -> Tuple[str, ...]:
        spam_score = content.get("score")
        if spam_score is not None:
            junk_score = "X" * (math.frexp(spam_score)[1] - 4)
        else:
            spam_score = "error"
            junk_score = "error"

        headers = [
            f"X-Spam-Score: {spam_score}",
            f"X-Spam-Threshold: {content.get('required_score', 'error')}",
            f"X-Junk-Score: {junk_score}",
            f"X-Spam-Action: {content.get('action') or ''}",
        ]

        symbols = content.get("symbols") or {}
        for index, value in enumerate(symbols.values()):
            headers.append(
                f"X-Spam-Symbol-{index + 1}: "
                f"{value['name']} ({value['score']}) {value.get('description') or ''}".strip()
            )
        return tuple(headers)

    async def check_file(
            self,
            from_: Optional[str],
            ip: Optional[str],
            rcpts: Optional[Collection[Optional[str]]],
            content_size: int,
            content_iterator: AsyncIterable,
            timeout: Optional[float] = None,
    ) -> Tuple[str, ...]:
        self._logger.debug("Checking email...")
        try:
            request_headers = (
                ("Accept-Encoding", "identity"),
                ("From", from_),
                ("Ip", ip),
                *(("Rcpt", rcpt) for rcpt in rcpts or ()),
            )

            status, headers, resp_body = await self._client_session.request(
                "POST", "/checkv2",
                request_headers,
                content_size,
                content_iterator,
                timeout,
            )
            self._logger.debug(f"Response: status={status!r} headers={headers} body={resp_body!r}")
        except ConnectionError as exc:
            raise RSPAMDConnectionError(f"Unable to connect to RsmapD: {str(exc)}") from exc
        except ResponseError as exc:
            raise RSPAMDConnectionError(f"RsmapD response error: {str(exc)}") from exc
        except asyncio.TimeoutError as exc:
            raise RSPAMDTimeoutError("Request to RspamD was terminated by timeout") from exc

        resp_json_data = json.loads(resp_body)
        if "error" in resp_json_data:
            self._logger.error(f"Rspamd error: {resp_json_data['error']}")
            raise RSPAMDResponseError(resp_json_data["error"])

        checking_result = self._prepare_check_file_result(resp_json_data)
        self._logger.debug(f"Checking email result: {checking_result}")
        return checking_result


# CGP client
############
class ProtocolMessage:
    def __init__(self, req_number: str, command: str, args: Sequence[Any] = ()) -> None:
        self._req_number = req_number
        self._command = command
        self._args = args

    @property
    def req_number(self) -> str:
        return self._req_number

    @property
    def command(self) -> str:
        return self._command

    @property
    def args(self) -> Sequence[Any]:
        return self._args

    def __repr__(self) -> str:
        return f"request_number={self.req_number} command={self.command} args={self.args}"


class Request(ProtocolMessage):
    pass


class Response(ProtocolMessage):
    pass


class CommandHandlerNotFound(Exception):
    pass


class _AsyncContentIterator:
    def __init__(self, opened_file: BinaryIO, chunk_size: int = 1024) -> None:
        self._chunk_size = chunk_size
        self._fd = opened_file

    def __aiter__(self) -> '_AsyncContentIterator':
        return self

    async def __anext__(self) -> bytes:
        chunk = self._fd.read(self._chunk_size)
        await asyncio.sleep(0)
        if chunk == b'':
            raise StopAsyncIteration()
        return chunk


class CommunigateMessage:
    _cgp_header_regexp = re.compile(r"^[A-Z]+ .*$")
    _headers_regexp = (
        (re.compile(r"^S.*\[([0-9a-f.:]+)\]$"), "ip"),
        (re.compile(r"^P.*<(.+)>$"), "from",),
        (re.compile(r"^R.*<(.+)>$"), "to",),
    )

    def __init__(self, file_path: str) -> None:
        self._file_path = file_path
        self._size = os.stat(file_path).st_size

        self._from: Optional[str] = None
        self._received_ip: Optional[str] = None
        self._to: Optional[Union[List[str], Tuple[None]]] = None
        self._content_offset: int = 0
        self._content_size: Optional[int] = None
        self._was_parsed = False

    @property
    def from_(self) -> Optional[str]:
        return self._from

    @property
    def received_ip(self) -> Optional[str]:
        return self._received_ip

    @property
    def to(self) -> Optional[Union[List[str], Tuple[None]]]:
        return self._to

    @property
    def content_size(self) -> Optional[int]:
        return self._content_size

    def _parse_message(self) -> None:
        headers: Dict[str, List[str]] = {}
        cgp_header = self._cgp_header_regexp
        headers_regexp = self._headers_regexp
        with open(self._file_path, "rb") as fd:
            while True:
                try:
                    file_line = fd.readline()
                    line = file_line.decode("utf8").strip("\r\n")
                except UnicodeDecodeError:
                    raise CommunigateMessageError("Failed to decode service information line")

                if not line:
                    break

                if not cgp_header.match(line):
                    # return cursor to start of this line
                    fd.seek(-len(file_line), 1)
                    break

                for header_value_regexp, header_name in headers_regexp:
                    header_values = header_value_regexp.findall(line)
                    if header_values:
                        headers.setdefault(header_name, []).append(header_values[0])

            self._content_offset = fd.tell()
            self._content_size = self._size - self._content_offset

        self._from = (headers.get("from") or (None,))[0]
        self._received_ip = (headers.get("ip") or (None,))[0]
        self._to = headers.get("to") or (None,)
        self._was_parsed = True

    @contextmanager
    def content_iterator(self) -> Generator[_AsyncContentIterator, None, None]:
        if not self._was_parsed:
            raise CommunigateMessageError("Message have not parsed")

        with open(self._file_path, "rb") as fd:
            fd.seek(self._content_offset)
            yield _AsyncContentIterator(fd)


async def parse_communigate_message(file_path: str) -> CommunigateMessage:
    message = CommunigateMessage(file_path)
    await get_event_loop().run_in_executor(None, message._parse_message)
    return message


class ExternalMessageFiltersManager:
    _logger = logging.getLogger("external_message_filters_manager")

    def __init__(self, base_path: str, rspam_session: RSPAMDService) -> None:
        self._base_path = base_path
        self._rspam_session = rspam_session
        self._commands = {
            "file": self._file_handler,
            "intf": self._protocol_version_handler
        }

    def get_request_handler(self, request: Request) -> Callable[[Request], Awaitable[Response]]:
        try:
            return self._commands[request.command.lower()]
        except KeyError:
            raise CommandHandlerNotFound

    @staticmethod
    async def _protocol_version_handler(request: Request) -> Response:
        return Response(request.req_number, "INTF", (4,))

    async def _file_handler(self, request: Request) -> Response:
        file_path = os.path.join(self._base_path, request.args[0])
        self._logger.debug(f"Message file: {file_path}")
        try:
            message = await parse_communigate_message(file_path)
        except CommunigateMessageError as exc:
            self._logger.error(f"Reading file service info error: {str(exc)}")
            return Response(request.req_number, "OK")

        content_size = message.content_size
        if content_size is None:
            self._logger.error("Reading file service info error: content_size is None")
            return Response(request.req_number, "OK")

        self._logger.debug("Message file parsed.")
        try:
            iterator: AsyncIterable[bytes]
            with message.content_iterator() as iterator:
                rspamd_check_result = await self._rspam_session.check_file(
                    message.from_, message.received_ip, message.to, content_size, iterator,
                )
        except (RSPAMDConnectionError, RSPAMDResponseError, RSPAMDTimeoutError) as exc:
            self._logger.error(f"Check file error: {str(exc)}")
            return Response(request.req_number, "OK")

        result = "\\e".join(rspamd_check_result)
        return Response(request.req_number, "ADDHEADER", (f"\"{result}\"", "OK"))


class RequestProtocol(asyncio.Protocol):
    _logger = logging.getLogger("communigate")

    def __init__(self, manager: ExternalMessageFiltersManager, writer: WriteTransport) -> None:
        self._manager = manager
        self._loop = get_event_loop()
        self._next_message: Optional[bytes] = None
        self._messages: Deque[Request] = deque()
        self._waiter: Optional[Future] = None

        self._writer = writer
        self._transport: Optional[BaseTransport] = None
        self._request_jobs: List[Job] = []
        self._protocol_job: Optional[Job] = None

    def connection_made(self, transport: BaseTransport) -> None:
        super().connection_made(transport)

        async def stop_callback(*_) -> None:
            self._loop.stop()
            self._logger.debug("Loop stopped")

        self._protocol_job = Job(self._start(), stop_callback)
        self._transport = transport
        self._logger.debug("Connection made")

    def data_received(self, data: bytes) -> None:
        self._logger.debug(f"Received data: {data!r}")
        if self._next_message:
            data = self._next_message + data
        self._logger.debug(f"Command data: {data!r}")

        data_split = data.split(b"\n")
        if len(data_split) > 1 and not data.endswith(b"\n"):
            self._next_message = data_split[-1]
            data_split = data_split[:-1]
        else:
            self._next_message = None

        for data in data_split:
            str_data = data.decode("utf8").strip()
            if str_data:
                split_data = str_data.split(" ")
                self._messages.append(Request(split_data[0], split_data[1], split_data[2:]))
                waiter = self._waiter
                if waiter is not None and not waiter.done():
                    waiter.set_result(None)

    def connection_lost(self, exception: Optional[Exception]) -> None:
        for job in self._request_jobs:
            job.cancel()
        self._writer.close()
        self._logger.debug("Connection lost")

    async def _get_request(self) -> Optional[Request]:
        if not self._messages:
            try:
                self._waiter = self._loop.create_future()
                self._logger.debug("Request waiting...")
                await self._waiter
            except asyncio.CancelledError:
                self._logger.warning("Request waiting cancelled")
                return None
            finally:
                self._waiter = None
        return self._messages.popleft()

    async def _write_response(self, resp: Response) -> None:
        msg = f"{resp.req_number} {resp.command} " \
            f"{(' ' + ' '.join(str(el) for el in resp.args)) if resp.args else ''}\r\n"
        self._logger.debug(f"Write response message: {msg}")
        self._writer.write(msg.encode("utf8"))

    async def _start(self) -> None:
        request_jobs = self._request_jobs

        async def finish_request(job, response) -> None:
            if response is None:
                response = Response(job.name, "OK")
            await self._write_response(response)
            request_jobs.remove(job)

        logger = self._logger
        while True:
            request = await self._get_request()
            if request is None:
                continue

            logger.debug(f"Received request: {request}")

            try:
                handler = self._manager.get_request_handler(request)
            except CommandHandlerNotFound:
                if request.command == "QUIT":
                    await self._write_response(Response(request.req_number, "OK"))
                    self._shutdown()
                    return
                logger.error(f"Request {request.req_number} unknown command: {request.command}")
                await self._write_response(Response(request.req_number, "OK"))
            else:
                request_jobs.append(Job(handler(request), finish_request, request.req_number))

    def _shutdown(self) -> None:
        for job in self._request_jobs:
            job.cancel()

        if self._transport:
            self._transport.close()

        self._writer.close()
        if self._protocol_job:
            self._protocol_job.cancel()
        self._logger.debug("Connection closed")


# main
######

class Formatter(logging.Formatter):
    def __init__(self, include_datetime: bool = False, prefix: str = "") -> None:
        fmt = "(%(name)s) %(message)s %(exc_info)s"
        if include_datetime:
            fmt = f"%(asctime)s {fmt}"

        if prefix:
            fmt = f"{prefix} {fmt}"

        super().__init__(fmt, None, "%")
        self._prefix = prefix
        self._include_datetime = include_datetime

    def format(self, record: logging.LogRecord) -> str:
        message = self.formatMessage(record)
        record.message = message
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            message = f"{message} {record.exc_text}"
        return message.replace("\n\r", "\t").replace("\n", "\t")

    def formatMessage(self, record: logging.LogRecord) -> str:
        parts: List[str] = []
        if self._prefix:
            parts.append(self._prefix)

        if self._include_datetime:
            parts.append(self.formatTime(record))

        req_number = get_current_ctx("background_job")
        if req_number:
            parts.append(f"[{req_number}]")

        parts.append(f"({record.name})")
        message = record.getMessage()
        parts.append(message)
        return " ".join(parts)


def init_loggers(base_dir: str, level: int) -> None:
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(Formatter(prefix="*"))
    logging.root.addHandler(stdout_handler)
    logging.root.setLevel(level)

    logs_dir = os.path.join(base_dir, "Logs")
    os.makedirs(logs_dir, exist_ok=True)
    file_handler = TimedRotatingFileHandler(
        os.path.join(logs_dir, "drweb-plugin.log"),
        when="d",
        backupCount=10,
    )
    file_handler.setFormatter(Formatter(include_datetime=True))
    file_handler.setLevel(level)
    logging.root.addHandler(file_handler)


class InitConnectorError(Exception):
    pass


async def start(args: argparse.Namespace) -> None:
    rspamd_connector: BaseConnector
    if args.rspamd_http_socket:
        rspamd_http_socket_list = args.rspamd_http_socket.split(":")
        rspamd_connector = TcpConnector(rspamd_http_socket_list[0], int(rspamd_http_socket_list[1]))
    elif args.rspamd_unix_socket:
        rspamd_connector = UnixConnector(args.rspamd_unix_socket)
    else:
        raise InitConnectorError("Rspamd connection type not set. Use --help to view rspamd connection types.")

    if args.debug:
        async def _run_forever() -> None:
            logger = logging.getLogger("ping")
            while True:
                await asyncio.sleep(5)
                logger.debug("I'm working...")
    else:
        async def _run_forever() -> None:
            while True:
                await asyncio.sleep(3600)

    fd = io.open(sys.stdin.fileno())
    try:
        loop = get_event_loop()
        async with ClientSession(connector=rspamd_connector, timeout=args.rspamd_request_timeout) as client_session:
            manager = ExternalMessageFiltersManager(args.base_dir, RSPAMDService(client_session))
            writer_transport, _ = await loop.connect_write_pipe(asyncio.Protocol, sys.stdout)
            request_protocol = RequestProtocol(manager, writer_transport)  # type: ignore
            await loop.connect_read_pipe(lambda: request_protocol, fd)

            await _run_forever()
    finally:
        fd.close()


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="Dr.Web CommuniGate Plugin")
    base_directory = "/var/CommuniGate"
    plugin_directory = os.path.join(base_directory, "DrWebPlugin")
    rspamd_request_timeout = 2 * 60  # 2 min
    parser.add_argument(
        "-v", "--version", action="version", version=f"DrWeb CommuniGate plugin version {__version__}",
        help="Show plugin version and exit"
    )
    parser.add_argument(
        "-d", "--base_dir", type=str, default=base_directory,
        help=f"CommuniGate base directory (default: {base_directory})",
    )
    parser.add_argument(
        "-p", "--plugin_directory", type=str, default=plugin_directory,
        help=f"Plugin files directory (default: {plugin_directory})"
    )
    parser.add_argument(
        "-r", "--rspamd_http_socket", type=str,
        help="RspamD http socket address"
    )
    parser.add_argument(
        "-u", "--rspamd_unix_socket", type=str,
        help="RspamD unix socket path"
    )
    parser.add_argument(
        "-t", "--rspamd_request_timeout", type=float, default=rspamd_request_timeout,
        help=f"RspamD request timeout (default: {rspamd_request_timeout} sec). Value must be set in seconds"
    )
    parser.add_argument("--debug", action="store_true", help="Enable debug mode")
    parser.set_defaults(debug=False)
    return parser.parse_args()


if __name__ == "__main__":
    sys.stderr.write = lambda *args, **kwargs: sys.stdout.write(*args, **kwargs)  # type: ignore
    args = parse_args()

    loop = get_event_loop()
    try:
        init_loggers(args.plugin_directory, logging.DEBUG if args.debug else logging.INFO)
        python_version = sys.version_info
        logging.info(
            f"DrWeb CommuniGate plugin version {__version__} "
            f"(python: {python_version.major}.{python_version.minor}.{python_version.micro}) started with args: "
            + ", ".join(f"{key}={val}" for key, val in vars(args).items() if val is not None)
        )

        if args.rspamd_unix_socket and args.rspamd_http_socket:
            logging.info(
                "Only one type of connection to rspamd must be specified. "
                "Do not use arguments 'rspamd_http_socket' and 'rspamd_unix_socket' at the same time."
            )
            exit(1)

        loop.run_until_complete(start(args))
    except KeyboardInterrupt:
        logging.error("KeyboardInterrupt")
        exit(1)
    except InitConnectorError as exc:
        logging.error(str(exc))
        exit(1)
    except Exception:
        logging.error("Exception:", exc_info=True)
        exit(1)
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()
